@@ -27,7 +27,7 @@ __copyright__ = "Copyright 2022, The Materials Project"
 __maintainer__ = "Jimmy-Xuan Shen @jmmshn"
 __date__ = "Mar 15, 2022"
 
-logger = logging.getLogger(__name__)
+_logger = logging.getLogger(__name__)
 
 
 class DefectType(Enum):
@@ -447,7 +447,7 @@ class Interstitial(Defect):
         # found in the ICSD
         inter_states = self.site.specie.icsd_oxidation_states[:2]
         if len(inter_states) == 0:
-            logger.warning(
+            _logger.warning(
                 f"No oxidation states found for {self.site.specie.symbol}. "
                 "in ICSD using `oxidation_states` without frequencuy ranking."
             )
