@@ -1,6 +1,7 @@
 """Utilities for defects module."""
 from __future__ import annotations
 
+import bisect
 import collections
 import logging
 import math
@@ -8,7 +9,7 @@ import operator
 from copy import deepcopy
 from functools import cached_property
 from pathlib import Path
-from typing import Generator
+from typing import Any, Callable, Generator
 
 import numpy as np
 from monty.json import MSONable
@@ -17,7 +18,8 @@ from numpy.linalg import norm
 from pymatgen.analysis.local_env import cn_opt_params
 from pymatgen.analysis.structure_matcher import StructureMatcher
 from pymatgen.core import Lattice, Structure
-from pymatgen.io.vasp import VolumetricData
+from pymatgen.electronic_structure.core import Spin
+from pymatgen.io.vasp.outputs import BandStructure, Procar, VolumetricData
 from pymatgen.io.vasp.sets import get_valid_magmom_struct
 from scipy.cluster.hierarchy import fcluster, linkage
 from scipy.spatial.distance import squareform
@@ -38,7 +40,7 @@ __email__ = "jmmshn@gmail.com"
 __status__ = "Development"
 __date__ = "Aug 15, 2022"
 
-logger = logging.getLogger(__name__)
+_logger = logging.getLogger(__name__)
 hart_to_ev = 27.2114
 ang_to_bohr = 1.8897
 invang_to_ev = 3.80986
@@ -202,22 +204,16 @@ def converge(f, step, tol, max_h):
 def get_zfile(
     directory: Path, base_name: str, allow_missing: bool = False
 ) -> Path | None:
-    """
-    Find gzipped or non-gzipped versions of a file in a directory listing.
+    """Find gzipped or non-gzipped versions of a file in a directory listing.
 
-    Parameters
-    ----------
-    directory : list of Path
-        A list of files in a directory.
-    base_name : str
-        The base name of file file to find.
-    allow_missing : bool
-        Whether to error if no version of the file (gzipped or un-gzipped) can be found.
-    Returns
-    -------
-    Path or None
-        A path to the matched file. If ``allow_missing=True`` and the file cannot be
-        found, then ``None`` will be returned.
+    Args:
+        directory : A list of files in a directory.
+        base_name : The base name of file file to find.
+        allow_missing : Whether to error if no version of the file (gzipped or un-gzipped) can be found.
+
+    Returns:
+        Path | None: A path to the matched file. If ``allow_missing=True``
+        and the file cannot be found, then ``None`` will be returned.
     """
     for file in directory.glob(f"{base_name}*"):
         if base_name == file.name:
@@ -234,13 +230,15 @@ def get_zfile(
 
 
 def generic_groupby(list_in, comp=operator.eq):
-    """
-    Group a list of unsortable objects
+    """Group a list of unsortable objects.
+
     Args:
         list_in: A list of generic objects
         comp: (Default value = operator.eq) The comparator
+
     Returns:
-        [int] list of labels for the input list
+        list[int]: list of labels for the input list
+
     """
     list_out = [None] * len(list_in)
     label_num = 0
@@ -260,9 +258,9 @@ def generic_groupby(list_in, comp=operator.eq):
 
 
 def get_local_extrema(chgcar: VolumetricData, find_min: bool = True) -> npt.NDArray:
-    """
-    Get all local extrema fractional coordinates in charge density,
-    searching for local minimum by default. Note that sites are NOT grouped
+    """Get all local extrema fractional coordinates in charge density.
+
+    Searching for local minimum by default. Note that sites are NOT grouped
     symmetrically.
 
     Args:
@@ -273,7 +271,6 @@ def get_local_extrema(chgcar: VolumetricData, find_min: bool = True) -> npt.NDAr
         extrema_coords (list): list of fractional coordinates corresponding
             to local extrema.
     """
-
     if find_min:
         sign = -1
     else:
@@ -298,8 +295,7 @@ def get_local_extrema(chgcar: VolumetricData, find_min: bool = True) -> npt.NDAr
 def remove_collisions(
     fcoords: npt.NDArray, structure: Structure, min_dist: float = 0.9
 ) -> npt.NDArray:
-    """
-    Removed points that are too close to existing atoms in the structure
+    """Removed points that are too close to existing atoms in the structure.
 
     Args:
         fcoords (npt.ArrayLike): fractional coordinates of points to remove
@@ -310,7 +306,7 @@ def remove_collisions(
         fcoord (numpy.ndarray): The filtered coordinates.
     """
     s_fcoord = structure.frac_coords
-    logger.info(s_fcoord)
+    _logger.info(s_fcoord)
     dist_matrix = structure.lattice.get_all_distances(fcoords, s_fcoord)
     all_dist = np.min(dist_matrix, axis=1)
     return np.array(
@@ -321,10 +317,11 @@ def remove_collisions(
 def cluster_nodes(
     fcoords: npt.ArrayLike, lattice: Lattice, tol: float = 0.2
 ) -> npt.NDArray:
-    """
-    Cluster nodes that are too close together using a tol.
+    """Cluster nodes that are too close together using hiercharcal clustering.
 
     Args:
+        fcoords (npt.ArrayLike): fractional coordinates of points to cluster.
+        lattice (Lattice): The lattice of the structure.
         tol (float): A distance tolerance. PBC is taken into account.
     """
     # Manually generate the distance matrix (which needs to take into
@@ -399,7 +396,8 @@ def get_avg_chg(
 
 
 class ChargeInsertionAnalyzer(MSONable):
-    """
+    """Object for insertions sites from charge density.
+
     Analyze the charge density and create new candidate structures by inserting at each charge minima
     The similar inserterd structures are given the same uniqueness label.
 
@@ -408,13 +406,22 @@ class ChargeInsertionAnalyzer(MSONable):
         often contains spurious local minima in the core. However you can still use CHGCAR
         with an appropriate ``max_avg_charge`` value.
 
-        Since the user might want to rerun their analysis with different ``avg_charge`` and ``max_avg_charge`` values,
-        we will generate and store all the ion-inserted structure and their uniqueness labels first and allow
-        the user to get the filtered and labeled results.
+        Since the user might want to rerun their analysis with different ``avg_charge`` and
+        ``max_avg_charge`` values, we will generate and store all the ion-inserted structure
+        and their uniqueness labels first and allow the user to get the filtered and labeled results.
 
         If you use this code please cite the following paper:
         J.-X. Shen et al.: npj Comput. Mater. 6, 1 (2020)
         https://www.nature.com/articles/s41524-020-00422-3
+
+    Attributes:
+        chgcar: The charge density object to analyze
+        working_ion: The working ion to be inserted
+        clustering_tol: Distance tolerance for grouping sites together
+        ltol: StructureMatcher ltol parameter
+        stol: StructureMatcher stol parameter
+        angle_tol: StructureMatcher angle_tol parameter
+        min_dist: Minimum distance between sites and the host atoms in Å.
     """
 
     def __init__(
@@ -427,16 +434,6 @@ class ChargeInsertionAnalyzer(MSONable):
         angle_tol: float = 5,
         min_dist: float = 0.9,
     ):
-        """
-        Args:
-            chgcar: The charge density object to analyze
-            working_ion: The working ion to be inserted
-            clustering_tol: Distance tolerance for grouping sites together
-            ltol: StructureMatcher ltol parameter
-            stol: StructureMatcher stol parameter
-            angle_tol: StructureMatcher angle_tol parameter
-            min_dist: Minimum distance between sites and the host atoms in Å.
-        """
         self.chgcar = chgcar
         self.working_ion = working_ion
         self.sm = StructureMatcher(ltol=ltol, stol=stol, angle_tol=angle_tol)
@@ -521,3 +518,124 @@ class ChargeInsertionAnalyzer(MSONable):
             res.append((avg_chg, [self.local_minima[idx] for idx in lab_groups[lab]]))
 
         return res
+
+
+def _get_ipr(spin, k_index, procar):
+    states = procar.data[spin][k_index, ...]
+    flat_states = states.reshape(states.shape[0], -1)
+    return 1 / np.sum(flat_states**2, axis=1)
+
+
+def get_ipr_in_window(
+    bandstructure: BandStructure,
+    procar: Procar,
+    band_window: int = 5,
+) -> dict[Spin, npt.NDArray]:
+    """Get the inverse participation ratio (IPR) of the states in a given band window.
+
+    For a given number of bands above and below the Fermi level, the IPR is calculated.
+
+    Args:
+        bandstructure: The bandstructure object.
+            The band just below the fermi level is used as the center of band window.
+        procar: The procar object.
+        k_index: The index of the k-point to use. If None, the IPR is averaged over all k-points.
+        band_range: The number of bands above and blow the fermi level to include in the search window.
+
+    Returns:
+        dict[Spin, npt.NDArray]: The IPR of the states in the band window.
+    """
+    res = dict()
+    for spin in bandstructure.bands.keys():
+        # last band that is fully below the fermi level
+        last_occ_idx = bisect.bisect_left(
+            bandstructure.bands[spin].max(1), bandstructure.efermi
+        )
+        lbound = max(last_occ_idx - band_window, 0)
+        ubound = min(last_occ_idx + band_window, bandstructure.nb_bands)
+        ib_ipr = np.array([])
+        for k_idx, _ in enumerate(bandstructure.kpoints):
+            ipr = _get_ipr(spin, k_idx, procar)
+            ib_ipr_ = np.stack((np.arange(lbound, ubound), ipr[lbound:ubound])).T
+            ib_ipr = np.stack((ib_ipr, ib_ipr_), axis=0) if ib_ipr.size else ib_ipr_
+        res[spin] = ib_ipr
+    return res
+
+
+def get_localized_state(
+    bandstructure: BandStructure,
+    procar: Procar,
+    k_index: int | None = None,
+    band_window: int = 5,
+) -> dict[Spin, tuple[int, float]]:
+    """Find the index of the localized state.
+
+    Find the band index with the lowest inverse participation ratio (IPR) in a small window
+    around the fermi level.  Since some of the core states can be very localized,
+    we should only search near the valence.
+
+    Args:
+        bandstructure: The bandstructure object.
+            The band just below the fermi level is used as the center of band window.
+        procar: The procar object.
+        k_index: The index of the k-point to use. If None, the IPR is averaged over all k-points.
+        band_range: The number of bands above and blow the fermi level to include in the search window.
+
+    Returns:
+        dict[Spin, tuple(int, float)]:
+            The band index and inverse participation ratio (IPR) of the most localized state for each spin channel.
+    """
+    d_ib_ipr = get_ipr_in_window(bandstructure, procar, band_window)
+    res = dict()
+    for spin, ib_ipr in d_ib_ipr.items():
+        # collape the first dimension by averaging over k-points
+        # This whole thing is not accurate enough to worry about k-weighting
+        if k_index is None:
+            ib_ipr = np.mean(ib_ipr, axis=0)
+            ib_ipr = np.expand_dims(ib_ipr, axis=0)
+            k_index = 0
+        # find the index of the minimum IPR
+        min_idx = np.argmin(ib_ipr[k_index, :, 1])
+        if not ib_ipr[k_index, min_idx, 0].is_integer():  # pragma: no cover
+            raise ValueError(
+                "Since the band indices are synced across "
+                "different k points this should always be an integer."
+            )
+        res[spin] = (int(ib_ipr[k_index, min_idx, 0]), ib_ipr[k_index, min_idx, 1])
+    return res
+
+
+def sort_positive_definite(
+    list_in: list, ref1: Any, ref2: Any, dist: Callable
+) -> tuple[list, list[float]]:
+    """Sort a list where we can only compute a positive-definite distance.
+
+    Sometimes, we can only compute a positive-definite distance between two objects.
+    (E.g., the displacement between two structures). In these cases, standard
+    sorting algorithms will not work. Here, we accept two reference points to give
+    some sense of direction. We then sort the list based on the distance between the
+    reference points. Note: this only works if the list falls on a line of some sort.
+
+    Args:
+        list_in: The list to sort.
+        ref1: The first reference point, this will be the `zero` point.
+        ref2: The second reference point, this will determine the direction.
+        dist: Some positive definite distance function.
+
+    Returns:
+        - the sorted list of objects.
+        - the signed distance in the chosen direction.
+    """
+    d1 = [dist(el, ref1) for el in list_in]
+    d2 = [dist(el, ref2) for el in list_in]
+    D0 = dist(ref1, ref2)
+
+    d_vs_s = []
+    for q1, q2, el in zip(d1, d2, list_in):
+        sign = +1
+        if q1 < q2 and q2 > D0:
+            sign = -1
+        d_vs_s.append((sign * q1, el))
+    d_vs_s.sort()
+    distances, sorted_list = list(zip(*d_vs_s))
+    return sorted_list, distances
